/**
  ******************************************************************************
  * @file           : nmea.c
  * @brief          : NMEA Driver
  * @author			: ELKYS servis s.r.o. (2019)
  ******************************************************************************
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
  *   1. Redistributions of source code must retain the above copyright notice,
  *      this list of conditions and the following disclaimer.
  *   2. Redistributions in binary form must reproduce the above copyright notice,
  *      this list of conditions and the following disclaimer in the documentation
  *      and/or other materials provided with the distribution.
  *   3. Neither the name of ELKYS servis s.r.o.  nor the names of its contributors
  *      may be used to endorse or promote products derived from this software
  *      without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************
  */
#include <string.h>
#include "main.h"
#include "board.h"
#include "stm32l4xx_hal.h"
#include "nmea.h"
#include "buffer.h"

#define MAX_NMEA_MESSAGES	32

GNSSInfo_TypeDef GNSSInfo;
BufferItem_TypeDef NMEA_Msg;
uint16_t NMEA_DataIndex = 0;
uint8_t NMEA_LineComplete = 1;
Buffer_TypeDef RXMSG;
uint8_t rxGNSS; /* receive data from GNSS */

void NMEA_Init(void)
{
	BUFFER_Create(&RXMSG, 20);
	GNSSInfo.Mode = GNSS_MODE_OFF;
	//GNSS_PWRON();	//not power on now default state
	HAL_UART_Receive_DMA(&GNSS_UART_BUS_HANDLER, &rxGNSS, 1);
}

/*
 * Check if GNSS is running and toggle power state.
 */
void NMEA_ToggleON(void)
{
	if (GNSSInfo.Mode == GNSS_MODE_OFF)
	{
		GNSSInfo.Mode = GNSS_MODE_TRACKING;
		GNSS_PWRON();
	}
	else
	{
		GNSSInfo.Mode = GNSS_MODE_OFF;
		GNSS_PWROFF();
	}
}


static uint32_t NMEA_atoi(char *p)
{
        uint32_t out = 0;

        while ((*p >= '0' && *p <= '9') || *p == '.')
        {
                if (*p == '.') {
                        p++;
                        continue;
                }
                out *= 10;
                out += *p - '0';
                p++;
        }

        return out;
}

static uint32_t NMEA_atoiA(char *p)
{
        uint32_t out = 0;

        while ((*p >= '0' && *p <= '9'))
        {
                out *= 10;
                out += *p - '0';
                p++;
        }

        return out;
}

void NMEA_Handler(void)
{
	BufferItem_TypeDef msg;
	while (BUFFER_Get(&RXMSG, &msg) != 0xFF)
	{
		NMEA_Decode(&msg);
	}
}

void NMEA_Decode(BufferItem_TypeDef* Msg)
{
	char *pBuff;
	char temp[8];

	pBuff = (char *)Msg->data;

	/*f_printf(&GNSSInfo.logfile, "%s",  pBuff);
	f_sync(&GNSSInfo.logfile);
	*/
	//printf("%s\n",  pBuff);
	//return;
	if(!strncmp(pBuff, "$GPGGA", 6) || !strncmp(pBuff, "$GNGGA", 6))
	{
		pBuff += 7;
		GNSSInfo.hour	= (pBuff[0] - '0') * 10 + pBuff[1] - '0';
		GNSSInfo.min	= (pBuff[2] - '0') * 10 + pBuff[3] - '0';
		GNSSInfo.sec	= (pBuff[4] - '0') * 10 + pBuff[5] - '0';

		//Latitude
		pBuff = strstr(pBuff, ",") + 1;
		memset(temp, 0, 8);
		memcpy(temp, &pBuff[0], 2);
		GNSSInfo.Latitude = (float)atof(temp);
		memset(temp, 0, 8);
		memcpy(temp, &pBuff[2], 8);
		GNSSInfo.Latitude += (float)atof(temp)/60.0;
		pBuff = strstr(pBuff, ",") + 1;
		if (pBuff[0] == 'S') GNSSInfo.Latitude *= -1;
		//if (pBuff[0] == 'N')  *(long*)&GNSSInfo.Latitude |= 0x80000000;
		  uint8_t SatsInView_GPS;
		  uint8_t SatsInView_GLS;
		printf("Sats(%d/%d): %f,", GNSSInfo.SatsInView_GPS, GNSSInfo.SatsInView_GLS, GNSSInfo.Latitude);

		//Longitude
		pBuff = strstr(pBuff, ",") + 1;
		memset(temp, 0, 8);
		memcpy(temp, &pBuff[0], 3);
		GNSSInfo.Longitude = (float)atof(temp);
		memset(temp, 0, 8);
		memcpy(temp, &pBuff[3], 8);
		GNSSInfo.Longitude += (float)atof(temp)/60.0;
		pBuff = strstr(pBuff, ",") + 1;
		if (pBuff[0] == 'W') GNSSInfo.Longitude *= -1;
		printf("%f\n", GNSSInfo.Longitude);

		//Valid
		pBuff = strstr(pBuff, ",") + 1;
		GNSSInfo.Fix = pBuff[0] - '0';
		if ((GNSSInfo.Fix == 1) || (GNSSInfo.Fix == 2) || (GNSSInfo.Fix == 6))
		{
			GNSSInfo.Mode = GNSS_MODE_FIXED;
		}
		else  GNSSInfo.Mode = GNSS_MODE_TRACKING;
		GNSSInfo.Fix_prev =GNSSInfo.Fix;	//Save previous state

		pBuff = strstr(pBuff, ",") + 1;
		GNSSInfo.SatsUsed = (pBuff[0] - '0') * 10 + pBuff[1] - '0';

		// HDOP
		pBuff = strstr(pBuff, ",") + 1;
		GNSSInfo.HDOP = NMEA_atoi(pBuff);

		//Altitude
		pBuff = strstr(pBuff, ",") + 1;
		GNSSInfo.Altitude = NMEA_atoi(pBuff);

	}
	else if(!strncmp(pBuff, "$GPRMC", 6) || !strncmp(pBuff, "$GNRMC", 6))
	{
		pBuff += 7;
		RTC_TimeTypeDef sTime;
		sTime.Hours   = (pBuff[0] - '0') * 10 + pBuff[1] - '0';
		sTime.Minutes = (pBuff[2] - '0') * 10 + pBuff[3] - '0';
		sTime.Seconds = (pBuff[4] - '0') * 10 + pBuff[5] - '0';
		sTime.DayLightSaving = RTC_DAYLIGHTSAVING_NONE;
		sTime.StoreOperation = RTC_STOREOPERATION_RESET;

		//validity - A-ok, V-invalid
		pBuff = strstr(pBuff, ",") + 1;
		if (pBuff[0] == 'V')
		{
			GNSSInfo.DataValid = 0;
			return;	//data is not valid
		}

		//current Latitude
		pBuff = strstr(pBuff, ",") + 1;

		//North/South
		pBuff = strstr(pBuff, ",") + 1;

		//current Longitude
		pBuff = strstr(pBuff, ",") + 1;

		//East/West
		pBuff = strstr(pBuff, ",") + 1;

		//Speed in knots
		pBuff = strstr(pBuff, ",") + 1;

		//True Course
		pBuff = strstr(pBuff, ",") + 1;

		//date Stamp ddmmyy
		pBuff = strstr(pBuff, ",") + 1;
		RTC_DateTypeDef sDate;
		sDate.Date   = (pBuff[0] - '0') * 10 + pBuff[1] - '0';
		sDate.Month = (pBuff[2] - '0') * 10 + pBuff[3] - '0';
		sDate.Year = ((pBuff[4] - '0') * 10 + pBuff[5] - '0');
		sDate.WeekDay = RTC_WEEKDAY_MONDAY;
		if (GNSSInfo.DataValid == 0)	//Previous State is invalid
		{
			RTC_SetTime(sTime);
			RTC_SetDate(sDate);
			printf("GNSS> RTC Date and Time was updated...\r\n");
		}
		GNSSInfo.DataValid = 1;
		//Magnetic Variantion
		pBuff = strstr(pBuff, ",") + 1;
		//Magnetic Variantion East/WEST
		pBuff = strstr(pBuff, ",") + 1;
	}

	//pBuff = (char *)Info->RxMessage;
	else if(!strncmp(pBuff, "$GPGSV", 6))
	{
		pBuff += 7;
		//Messages


		//Message Number
		pBuff = strstr(pBuff, ",") + 1;

		//SatsInView
		pBuff = strstr(pBuff, ",") + 1;
		GNSSInfo.SatsInView_GPS = (pBuff[0] - '0') * 10 + pBuff[1] - '0';
	}
	else if(!strncmp(pBuff, "$GLGSV", 6))
	{
		pBuff += 7;
		//Messages


		//Message Number
		pBuff = strstr(pBuff, ",") + 1;

		//SatsInView
		pBuff = strstr(pBuff, ",") + 1;
		GNSSInfo.SatsInView_GLS = (pBuff[0] - '0') * 10 + pBuff[1] - '0';
	}
	//Example: $GPVTG,83.37,T,,M,0.00,N,0.0,K,A*32<CR><LF>
	else if(!strncmp(pBuff, "$GPVTG", 6) || !strncmp(pBuff, "$GNVTG", 6))
	{
		pBuff += 7;
		//Course True

		//Reference
		pBuff = strstr(pBuff, ",") + 1;

		//Course Magnetic
		pBuff = strstr(pBuff, ",") + 1;

		//Reference
		pBuff = strstr(pBuff, ",") + 1;

		//Speed (knots)
		pBuff = strstr(pBuff, ",") + 1;

		//Units
		pBuff = strstr(pBuff, ",") + 1;

		//Speed (km/h)
		pBuff = strstr(pBuff, ",") + 1;
		GNSSInfo.Speed_A = NMEA_atoiA(pBuff);
		pBuff = strstr(pBuff, ".") + 1;
		GNSSInfo.Speed_B = NMEA_atoi(pBuff);

		//Units
		pBuff = strstr(pBuff, ",") + 1;
	}
}
void NMEA_StoreByte(uint8_t* ucByte)
{
	if (*ucByte == NMEA_FLAG)
	{
		NMEA_DataIndex = 0;
		NMEA_LineComplete = 0;
	}
	if ((*ucByte == '\n') && (NMEA_Msg.data[NMEA_DataIndex-1] == '\r'))	//<CR><LF> detected
	{
		NMEA_Msg.data[NMEA_DataIndex++] = '\0';
		BUFFER_Add(&RXMSG, &NMEA_Msg);
		NMEA_LineComplete = 1;
	}
	else NMEA_Msg.data[NMEA_DataIndex++] = *ucByte;
	//printf("%c",*ucByte);
}
